@@ -31,19 +31,25 @@ our @EXPORT = qw(
     
 );
 
-our $VERSION = '0.04';
+our $VERSION = '0.08';
 
 # Preloaded methods go here.
 use X12::Parser::Cf;
+
+my $line_count        = 0;
 my $current_level     = undef;
+my $track_level       = undef;
 my $element_seperator = undef;
 my $segment_seperator = undef;
 
 my $cf       = undef;
-my $where    = -1;
-my $size     = -1;
-my @array    = ();
-my $analysis = undef;
+
+my $file_handle = undef;
+my @array_of_handles = ();
+
+my $last_read_segment = undef;
+my $last_read_loop = undef;
+
 
 
 sub new {
@@ -60,45 +66,31 @@ sub parse {
     my %params = @_;
     $self->{file} = $params{file};
     $self->{conf} = $params{conf};
-    $self->_load;
-    for ( my $i=0; $i<$size; $i++ ) {
-        vec ( $analysis, $i, 16 ) = 0xFFFF;
-    }
-    $current_level = 1;
-    my @level_one = $cf->get_level_one;
-    $self->_parse ( 0, \@level_one );
-}
-
-
-sub _load {
-    my $self = shift;
-
-    my $orig_line_seperator = undef;
-    my $isa            = undef;
 
     $cf  = new X12::Parser::Cf;
     $cf->load ("$self->{conf}") if defined $self->{conf};
+    
+    $track_level = 1;
+    my @level_one = $cf->get_level_one;
+    unshift ( @array_of_handles, \@level_one );
 
-    open (FILE, "$self->{file}") || die "error: cannot open file $self->{file}\n";
-    if ( read(FILE, $isa, 106) != 106 ) {
-        close (FILE);
+    open ($file_handle, "$self->{file}") || die "error: cannot open file $self->{file}\n";
+    $self->_set_seperator;
+}
+
+
+sub _set_seperator {
+    my $self = shift;
+    my $isa  = undef;
+
+    if ( read($file_handle, $isa, 106) != 106 ) {
+        close ($file_handle);
         die "error: invalid file format $self->{file}\n";
     }
-
-    $orig_line_seperator = $/;
+    seek ($file_handle, -106, 1);
 
     $segment_seperator = substr ($isa, 105, 1);
     $element_seperator = substr ($isa, 3, 1);
-
-    $/ = $segment_seperator;
-
-    seek (FILE, 0, 0);
-    @array = <FILE>;
-    close (FILE);
-    chop(@array);
-    $size = @array;
-
-    $/ = $orig_line_seperator;
 }
 
 
@@ -114,178 +106,236 @@ sub get_element_seperator {
 }
 
 
-sub _parse {
+sub _parse_loop_start {
     my $self = shift;
-    my $array_pos = shift;
-    my @array_of_handles = @_;
-    my ($segment, @segment );
+    my ($current_loop, $segment, @segment );
 
-    RECURSION:
-    for (my $i=$array_pos; $i<$size; $i++) {
-        $segment = $array[$i];
+    if ( defined $last_read_loop ) {
+            my $temp = $last_read_loop;
+            $last_read_loop = undef;
+            return $temp;
+    }
+
+    while ( $segment = <$file_handle> ) {
+        chomp($segment);
+        $line_count++;    
+        $last_read_segment = $segment;
         @segment = split ( /\Q$element_seperator\E/, $segment );
 
         for my $level_handle (@array_of_handles) {
-            foreach my $pos (@$level_handle) {
-                my $loop = $cf->{looptree}->[$pos][1];
+            foreach my $tree_index (@$level_handle) {
+                my $loop = $cf->{looptree}->[$tree_index][1];
                 my @left = split ( /:/, $cf->{segmentstart}->{$loop}->[0] );
                 if ( $left[0] eq $segment[0] ) {
                     if ( $left[2] eq "" ) {
-                        vec ( $analysis, $i, 16 ) = $pos;
-                        my $diff = $current_level - $cf->{looptree}->[$pos][0];
-                        $current_level =  $cf->{looptree}->[$pos][0];
+                        $current_loop = $cf->{looptree}->[$tree_index][1];
+                        $current_level = $cf->{looptree}->[$tree_index][0];
+                        my $diff = $track_level - $cf->{looptree}->[$tree_index][0];
+                        $track_level = $cf->{looptree}->[$tree_index][0];
                         while ( $diff > 0 ) {
                             shift ( @array_of_handles );
                             $diff--;
                         }
-                        my @next_level = $cf->get_next_level ( $pos );
+                        my @next_level = $cf->get_next_level ( $tree_index );
                         if ( 'END' ne $next_level[0] ) {
-                            $current_level++;
-                            my $ref = \@next_level;
-                            unshift ( @array_of_handles, $ref );
-                            $array_pos = $i + 1;
-                            next RECURSION; # recursion
+                            $track_level++;
+                            unshift ( @array_of_handles, \@next_level );
+                            return $current_loop;
                         }
-                        next RECURSION; # already matched goto next line
+                        return $current_loop;
                     }
                     else {
                         my @qual = split ( /,/, $left[2] );
                         if (( grep { $_ eq $segment[$left[1]] } @qual )) {
-                            vec ( $analysis, $i, 16 ) = $pos;
-                            my $diff = $current_level - $cf->{looptree}->[$pos][0];
-                            $current_level =  $cf->{looptree}->[$pos][0];
+                            $current_loop = $cf->{looptree}->[$tree_index][1];
+                            $current_level = $cf->{looptree}->[$tree_index][0];
+                            my $diff = $track_level - $cf->{looptree}->[$tree_index][0];
+                            $track_level = $cf->{looptree}->[$tree_index][0];
                             while ( $diff > 0 ) {
                                 shift ( @array_of_handles );
                                 $diff--;
                             }
-                            my @next_level = $cf->get_next_level ( $pos );
+                            my @next_level = $cf->get_next_level ( $tree_index );
                             if ( 'END' ne $next_level[0] ) {
-                                $current_level++;
-                                my $ref = \@next_level;
-                                unshift ( @array_of_handles, $ref );
-                                $array_pos = $i + 1;
-                                next RECURSION; # recursion
+                                $track_level++;
+                                unshift ( @array_of_handles, \@next_level );
+                                return $current_loop;
                             }
-                            next RECURSION; # already matched goto next line
+                            return $current_loop;
                         }
                     }
                 }
             }
         }
     }
+    close ($file_handle);
     return;
 }
 
 
-sub get_next_loop {
+sub _parse_loop {
     my $self = shift;
-    my $i    = $where;
+    my ( $segment, @segment, @loop );
 
-    for ( $i++ ;$i<$size; $i++ ) {
-        if ( vec ( $analysis, $i, 16 ) != 0xFFFF ) {
-            $where = $i;
-            return ( $cf->{looptree}->[vec($analysis,$i,16)]->[1] );
+    if ( defined $last_read_segment ) {
+            push (@loop, $last_read_segment);
+            $last_read_segment = undef;
+    }
+
+    while ( $segment = <$file_handle> ) {
+        chomp($segment);
+        $line_count++;    
+        $last_read_segment = $segment;
+        @segment = split ( /\Q$element_seperator\E/, $segment );
+
+        for my $level_handle (@array_of_handles) {
+            foreach my $tree_index (@$level_handle) {
+                my $loop = $cf->{looptree}->[$tree_index][1];
+                my @left = split ( /:/, $cf->{segmentstart}->{$loop}->[0] );
+                if ( $left[0] eq $segment[0] ) {
+                    if ( $left[2] eq "" ) {
+                        $last_read_loop = $cf->{looptree}->[$tree_index][1];
+                        $current_level = $cf->{looptree}->[$tree_index][0];
+                        my $diff = $track_level - $cf->{looptree}->[$tree_index][0];
+                        $track_level =  $cf->{looptree}->[$tree_index][0];
+                        while ( $diff > 0 ) {
+                            shift ( @array_of_handles );
+                            $diff--;
+                        }
+                        my @next_level = $cf->get_next_level ( $tree_index );
+                        if ( 'END' ne $next_level[0] ) {
+                            $track_level++;
+                            unshift ( @array_of_handles, \@next_level );
+                            return @loop;
+                        }
+                        return @loop;
+                    }
+                    else {
+                        my @qual = split ( /,/, $left[2] );
+                        if (( grep { $_ eq $segment[$left[1]] } @qual )) {
+                            $last_read_loop = $cf->{looptree}->[$tree_index][1];
+                            $current_level = $cf->{looptree}->[$tree_index][0];
+                            my $diff = $track_level - $cf->{looptree}->[$tree_index][0];
+                            $track_level =  $cf->{looptree}->[$tree_index][0];
+                            while ( $diff > 0 ) {
+                                shift ( @array_of_handles );
+                                $diff--;
+                            }
+                            my @next_level = $cf->get_next_level ( $tree_index );
+                            if ( 'END' ne $next_level[0] ) {
+                                $track_level++;
+                                unshift ( @array_of_handles, \@next_level );
+                                return @loop;
+                            }
+                            return @loop;;
+                        }
+                    }
+                }
+            }
         }
+        push (@loop, $segment);
     }
-    if ( $i >= $size ) {
-        return 0;
-    }
+    return @loop;
 }
 
 
-sub get_loop_segments {
-    my $self = shift;
-    my $loop = shift;
-    my @temp = ();
-    my $i    = $where;
 
-    if ( $cf->{looptree}->[vec ( $analysis, $i, 16 )]->[1] ne $loop ) {
-        return @temp;
+sub get_next_loop {
+    my $self = shift;
+    my $loop = undef;
+
+    my $original_line_seperator = $/;
+    $/ = $segment_seperator;
+
+    $loop = $self->_parse_loop_start;
+
+    $/ = $original_line_seperator;
+
+    if ( 'IEA' eq $loop ) {
+        if ( ! eof($file_handle) ) {
+            $self->_set_seperator;
+        }
     }
-    do {
-        push ( @temp, $array[$i] );
-    } while ( vec ( $analysis, ++$i, 16 ) == 0xFFFF && $i < $size );
-    return @temp;
+    return $loop;
 }
 
 
 sub get_next_pos_loop {
     my $self = shift;
-    my $i    = $where;
+    my $loop = undef;
 
-    for ( $i++ ;$i < $size; $i++ ) {
-        if ( vec ( $analysis, $i, 16 ) != 0xFFFF ) {
-            $where = $i;
-            return ( $i, $cf->{looptree}->[vec($analysis,$i,16)]->[1] );
+    my $original_line_seperator = $/;
+    $/ = $segment_seperator;
+
+    $loop = $self->_parse_loop_start;
+
+    $/ = $original_line_seperator;
+
+    if ( 'IEA' eq $loop ) {
+        if ( ! eof($file_handle) ) {
+            $self->_set_seperator;
         }
     }
-    if ( $i >= $size ) {
-        return;
-    }
+    if ( ! defined $loop ) { return }
+    return ($line_count, $loop);
 }
 
 
 sub get_next_pos_level_loop {
     my $self = shift;
-    my $i    = $where;
+    my $loop = undef;
 
-    for ( $i++ ;$i < $size; $i++ ) {
-        if ( vec ( $analysis, $i, 16 ) != 0xFFFF ) {
-            $where = $i;
-            return ( $i, $cf->{looptree}->[vec($analysis,$i,16)]->[0],
-                         $cf->{looptree}->[vec($analysis,$i,16)]->[1] );
+    my $original_line_seperator = $/;
+    $/ = $segment_seperator;
+
+    $loop = $self->_parse_loop_start;
+
+    $/ = $original_line_seperator;
+
+    if ( 'IEA' eq $loop ) {
+        if ( ! eof($file_handle) ) {
+            $self->_set_seperator;
         }
     }
-    if ( $i >= $size ) {
-        return;
-    }
+    if ( ! defined $loop ) { return }
+    return ($line_count, $current_level, $loop);
 }
 
 
-sub get_segments {
+sub get_loop_segments {
     my $self = shift;
-    my $i    = shift;
-    my @temp = ();
+    my @loop = ();
 
-    if ( $i < 0 || $i >= $size ) {
-        return @temp;
-    }
-    do {
-        push ( @temp, $array[$i] );
-    } while ( vec ( $analysis, ++$i, 16 ) == 0xFFFF && $i < $size );
-    return @temp;
-}
+    my $original_line_seperator = $/;
+    $/ = $segment_seperator;
 
+    @loop = $self->_parse_loop;
 
-sub reset_pos {
-    my $self = shift;
-    $where = -1;
+    $/ = $original_line_seperator;
+
+    return @loop;
 }
 
 
 sub print_tree {
     my $self  = shift;
-    my $index = undef;
-    my $pad   = undef;
-    my $level = undef;
+    my ($pad, $index, $segment);
 
-    for (my $i=0; $i<$size; ) {
-        $level = $cf->{looptree}->[vec($analysis,$i,16)]->[0];
+    while ( my ($pos, $level, $loop) = $self->get_next_pos_level_loop ) {
         if ( $level != 0 ) {
             $pad = '  |' x $level; 
-            print "       $pad--$cf->{looptree}->[vec($analysis,$i,16)]->[1]\n";
+            print "       $pad--$loop\n";
             $pad = '  |' x ( $level + 1 ); 
-            do {
-                $index = sprintf ( "%+7s", $i+1 );
-                print "$index$pad-- $array[$i]\n";
-            } while ( vec ( $analysis, ++$i, 16 ) == 0xFFFF && $i < $size );
-        }
-        else {
-            $i++;
+            my @loop = $self->get_loop_segments;
+            foreach $segment (@loop) {
+                $index = sprintf ( "%+7s", $pos++ );
+                print "$index$pad-- $segment\n";
+            }
         }
     }
 }
 
+
 1;
 __END__
 # Below is stub documentation for your module. You'd better edit it!
@@ -296,38 +346,38 @@ X12::Parser - Perl module for parsing X12 Transaction files
 
 =head1 SYNOPSIS
 
-use X12::Parser;
+  use X12::Parser;
 
-# Create a parser object
- my $p = new X12::Parser;
+  # Create a parser object
+  my $p = new X12::Parser;
 
-# Parse a file with the transaction specific configuration file
- $p->parse ( file => '837.txt',
-             conf => 'X12-837P.cf' 
-           );
+  # Parse a file with the transaction specific configuration file
+  $p->parse ( file => '837.txt',
+              conf => 'X12-837P.cf' 
+            );
 
-# Step through the file 
- while ( my $loop = $p->get_next_loop ) {
-   my @loop = $p->get_loop_segments ($loop); 
- }
+  # Step through the file 
+  while ( my $loop = $p->get_next_loop ) {
+    my @loop = $p->get_loop_segments; 
+  }
 
-# or use this method instead 
- while ( my ($pos, $loop) = $p->get_next_pos_loop ) { 
-   my @loop = $p->get_segments ($pos); 
- }
+  # or use this method instead 
+  while ( my ($pos, $loop) = $p->get_next_pos_loop ) { 
+    my @loop = $p->get_loop_segments; 
+  }
 
-# or use
- while ( my ($pos, $level, $loop) = $p->get_next_pos_level_loop ) {
-   my @loop = $p->get_segments ($pos);
- }
+  # or use
+  while ( my ($pos, $level, $loop) = $p->get_next_pos_level_loop ) {
+    my @loop = $p->get_loop_segments; 
+  }
 
 
 =head1 ABSTRACT
 
 X12::Parser package provides a efficient way to parse X12
-transaction files. Although this packages is built keeping HIPAA
+transaction files. Although this package is built keeping HIPAA
 related X12 transactions in mind, it is flexible and can be
-adapted to any X12 or similar transactions.
+adapted to parse any X12 or similar transactions.
 
 =head1 DESCRIPTION
 
@@ -343,13 +393,13 @@ $p = new X12::Parser;
    arguments. It only initializes the members variables required
    for parsing the transaction file.
 
-$p->parse ( file => '837.txt', conf => 'X12-837P.cf' );
+$p->parse(file => '837.txt', conf => 'X12-837P.cf');
    This method takes two arguments. The first argument is the 
-   transaction file which needs to be parsed. The second argument 
+   transaction file that needs to be parsed. The second argument 
    specifies the configuration file to be used for parsing the 
    transaction file. 
 
-   This package is a generic parser for parsing files which use a
+   This package is a generic parser for parsing files that use a
    format similar to the X12 specification. The ability to parse
    different transaction types is provided by means of using different
    configuration files. The configuration files for all the X12 HIPAA
@@ -358,44 +408,33 @@ $p->parse ( file => '837.txt', conf => 'X12-837P.cf' );
    To create your own configuration file read the X12::Parser::Readme 
    man page.
 
-$p->get_next_loop ();
-   This function returns the next loop name for the transaction file
-   which is being parsed. The loop name is as specified in the cf
+$p->get_next_loop;
+   This function returns the name of the next loop that is present
+   in the file being parsed. The loop name is as specified in the cf
    file.
 
-$p->get_loop_segments ($loop);
-   Pass the loop name returned by get_next_loop() to obtain the 
-   segments in this loop. This function is to be used in tandem with 
+$p->get_loop_segments;
+   This function returns the segments in the loop that was returned
+   by get_next_loop(). This function is to be used in tandem with 
    the get_next_loop. If not it may return/produce undesired results. 
-   get_next_loop() and get_loop_segments() internally keep record of 
-   the current loop position, so get_loop_segments will return the 
-   current loop segments even if you do not pass the loop name.
 
-$p->get_next_pos_loop ();
-   This function returns the next loop name and the loop position.
-   Note 0 corresponds to the first segment. 
+$p->get_next_pos_loop;
+   This function returns the next loop name and the segment position.
+   Note position 1 corresponds to the first segment. 
 
-$p->get_next_pos_level_loop ();
+$p->get_next_pos_level_loop;
    Same as get_next_pos_loop() except that in addition this function
    returns the level of the loop. The level corresponds to the level
    of the loop in the loop hierarchy. The top level loop has level 1.
 
-$p->get_segments ($pos);
-   get_segments returns an array of segments in the loop starting at
-   position $pos. If the specified position does not correspond to a 
-   loop, the function returns the segment at position $pos. 
-
-$p->reset_pos ();
-   Resets the pointer used by get_next_* functions. This causes the
-   get_next_* functions to start parsing again from the begining of
-   the file.
-
 $p->print_tree;
    Prints the transaction file in a tree format.
 
-$p->get_segment_seperator () 
-$p->get_element_seperator ()
-   Get the segment seperator, element seperator.
+$p->get_segment_seperator;
+   Get the segment seperator.
+
+$p->get_element_seperator;
+   Get the element seperator.
 
 The configuration files provided with this package and the corresponding
 transaction type is mentioned below. These are the X12 HIPAA transactions.
@@ -419,7 +458,7 @@ These cf files are installed in under the X12/Parser/cf folder.
 
 The sample cf files provided with this package are good to the best of
 the authors knowledge. However the user should ensure the validity of
-these files. The user may use them as is at his own risk.
+these files. The user may use them as is at their own risk.
 
 =head2 EXPORT
 
